@@ -17,24 +17,35 @@ import re
 import sys
 import errno
 import locale
+import inspect
+import logging
 import itertools
 import collections
 
 __author__  = 'Miguel Turner'
-__version__ = '0.3.1'
+__version__ = '0.3.2'
 __license__ = 'MIT'
 
 __all__ = ['Config', 'IniFormat', 'ConfigError', 'Coercer', 'CoerceError']
 
 PY3 = sys.version_info.major >= 3
-
 # use str for unicode data and bytes for binary data
 if not PY3:
     str = unicode
 
+WIN = os.name == 'nt'
+if WIN:
+    import ntpath
+    try:
+        import winreg
+    except ImportError:
+        import _winreg as winreg
+
 # the name *type* is used often so give type() an alias rather than use *typ*
 _type = type
 
+log = logging.getLogger('profig')
+
 ## Config ##
 
 class ConfigSection(collections.MutableMapping):
@@ -50,9 +61,9 @@ class ConfigSection(collections.MutableMapping):
         self._default = NoValue
         self._type = None
         self._parent = parent
+        self._dirty = False
         
         self.comment = None
-        self.dirty = False
         
         if parent is None:
             # root
@@ -98,6 +109,11 @@ class ConfigSection(collections.MutableMapping):
         """`True` if this section has a valid value. Read-only."""
         return not (self._value is NoValue and self._default is NoValue)
     
+    @property
+    def dirty(self):
+        """`True` if this section's value has changed since the last write. Read-only."""
+        return self._dirty
+    
     @property
     def is_default(self):
         return (self._value is NoValue and self._default is not NoValue)
@@ -161,22 +177,19 @@ class ConfigSection(collections.MutableMapping):
         file in a manner consistent with the active :class:`~profig.Format`.
         """
         section = self._create_section(key)
-        section._type = type or default.__class__
+        section._type = type or _type(default)
         section.set_default(default)
         section.comment = comment
     
-    def get(self, key, default=None, convert=True, type=None):
-        """
-        Return the value for key if key is in the dictionary,
-        else default. If *default* is not given, it defaults to
-        `None`, so that this method never raises an
-        :exc:`~profig.InvalidSectionError`. If *type* is provided,
-        it will be used as the type to convert the value from text.
-        If *convert* is `False`, *type* will be ignored.
+    def get(self, key, default=None):
+        """If *key* exists, returns the value. Otherwise, returns *default*.
+        
+        If *default* is not given, it defaults to `None`, so that this
+        method never raises an exception.
         """
         try:
             section = self.section(key, create=False)
-            return section.value(convert, type)
+            return section.value()
         except (InvalidSectionError, NoValueError):
             return default
     
@@ -195,10 +208,8 @@ class ConfigSection(collections.MutableMapping):
         del section._parent._children[section.name]
     
     def __bool__(self):
-        return True
-    
-    def __nonzero__(self):
-        return True
+        return self.valid or len(self) > 0
+    __nonzero__ = __bool__
     
     def __len__(self):
         return len(self._children)
@@ -223,37 +234,33 @@ class ConfigSection(collections.MutableMapping):
         return "{}('{}', value={!r}, keys={}, comment={!r})".format(
             self.__class__.__name__, self.key, value, list(self), self.comment)
     
-    def as_dict(self, flat=False, recurse=True, convert=True, dict_type=None):
+    def as_dict(self, flat=False, recurse=True, dict_type=None):
         """Returns the configuration's keys and values as a dictionary.
         
         If *flat* is `True`, returns a single-depth dict with :samp:`.`
         delimited keys.
         
-        If *convert* is `True`, all values will be converted. Otherwise, their
-        string representations will be returned.
-        
         If *dict_type* is not `None`, it should be the mapping class to use
         for the result. Otherwise, the *dict_type* set by
-        :meth:`~config.Config.__init__` will be used (the default is
-        `OrderedDict`).
+        :meth:`~config.Config.__init__` will be used.
         """
         dtype = dict_type or self._root._dict_type
         valid = self is not self._root and self.valid
         
         if flat:
             sections = ((k, self.section(k)) for k in self)
-            return dtype((k, s.value(convert)) for k, s in sections)
+            return dtype((k, s.value()) for k, s in sections)
         
         if recurse and self._children:
             d = dtype()
             if valid:
-                d[''] = self.value(convert)
+                d[''] = self.value()
             for section in self.sections():
-                d.update(section.as_dict(convert=convert, dict_type=dict_type))
+                d.update(section.as_dict(dict_type=dict_type))
             
             return d if self is self._root else dtype({self.name: d})
         elif valid:
-            return dtype({self.name: self.value(convert)})
+            return dtype({self.name: self.value()})
         else:
             return dtype()
 
@@ -275,136 +282,106 @@ class ConfigSection(collections.MutableMapping):
                 raise InvalidSectionError(key)
         return section
 
-    def sections(self, recurse=False):
+    def sections(self, recurse=False, only_valid=False):
         """Returns the sections that are children to this section.
         
         If *recurse* is `True`, returns grandchildren as well.
+        If *only_valid* is `True`, returns only valid sections.
         """
         for child in self._children.values():
-            yield child
+            if not only_valid or child.valid:
+                yield child
             if recurse:
                 for grand in child.sections(recurse):
-                    yield grand
+                    if not only_valid or grand.valid:
+                        yield grand
     
-    def reset(self, recurse=True):
+    def reset(self, recurse=True, clean=False):
         """Resets this section to it's default value, leaving it
         in the same state as after a call to :meth:`ConfigSection.init`.
-        If *recurse* is `True`, does the same to all the
-        section's children."""
+        
+        If *recurse* is `True`, does the same to all the section's children.
+        If *clean* is `True`, also clears the dirty flag on all sections.
+        """
         if self is not self._root:
-            self._reset()
+            self._reset(clean)
         if recurse:
             for section in self.sections(recurse):
-                section._reset()
+                section._reset(clean)
     
-    def value(self, convert=True, type=None):
-        """Get the section's value.
-        
-        To get the internal string value, set *convert* to `False`.
-        If *convert* is `False`, *type* will be ignored.
-        """
-        if not convert:
-            if self._value is not NoValue:
-                return self._value
-        elif self._value is not NoValue:
-            return self.convert(self._value, type)
-        
-        return self.default(convert, type)
+    def value(self):
+        """Get the section's value."""
+        if self._value is not NoValue:
+            return self._value
+        return self.default()
     
-    def set_value(self, value, adapt=True):
-        """Set the section's value.
-        
-        To set the internal string value, set *adapt* to `False`.
-        """
-        if adapt:
-            value = self.adapt(value)
-        elif isinstance(value, bytes) and self._type is not bytes:
-            value = value.decode(self._root.encoding)
-        if value != self._value:
-            self._value = value
-            self.dirty = True
+    def set_value(self, value):
+        """Set the section's value."""
+        self._value = value
+        self._dirty = True
     
-    def default(self, convert=True, type=None):
-        """Get the section's default value.
-        
-        To get the underlying string value, set *convert* to `False`.
-        If *convert* is `False`, *type* will be ignored.
-        """
-        if not convert:
-            if self._default is not NoValue:
-                return self._default
-        elif self._default is not NoValue:
-            return self.convert(self._default, type)
-        
+    def default(self):
+        """Get the section's default value."""
+        if self._default is not NoValue:
+            return self._default
         raise NoValueError(self.key)
     
-    def set_default(self, value, adapt=True):
-        """Set the section's default value.
-        
-        To set the internal string value, set *adapt* to `False`.
-        """
-        if adapt:
-            value = self.adapt(value)
-        elif isinstance(value, bytes) and self._type is not bytes:
-            value = value.decode(self._root.encoding)
+    def set_default(self, value):
+        """Set the section's default value."""
         self._default = value
     
-    def items(self, convert=True):
-        """Returns a (key, value) iterator over the unprocessed values of
-        this section."""
-        for key in self:
-            yield (key, self.section(key).value(convert))
-    
-    def adapt(self, value, type=None):
-        """
-        value -> str
-        """
+    def adapt(self):
+        """value -> str"""
         if not self._root.coercer:
             return value
-        # only set the type if it has not already been set
-        if self._type is None:
-            self._type = type or value.__class__
-        type = type or self._type
-        return self._root.coercer.adapt(value, type)
+        return self._root.coercer.adapt(self.value(), self._type)
     
-    def convert(self, string, type=None):
-        """
-        str -> value
-        """
+    def convert(self, string):
+        """str -> value"""
         if not self._root.coercer:
-            return string
-        type = type or self._type
-        return self._root.coercer.convert(string, type)
+            value = string
+        else:
+            type = self._type
+            if not (inspect.isclass(type) and issubclass(type, bytes)):
+                string = string.decode(self._root.encoding)
+            value = self._root.coercer.convert(string, type)
+        self.set_value(value)
     
     ## utilities ##
     
     def _read(self, sources, format):
         write_lines = None
+        # True if at least one source read something
+        one_source_read = False
         
         # read unchanged values from sources
         for i, source in enumerate(reversed(sources)):
             try:
                 file = format.open(source)
-            except IOError:
+            except IOError as e:
+                log.debug('%s: %s', source, e)
                 continue
             
             # read file
             try:
                 lines = format.read(file)
-            except IOError:
-                # XXX: there should be a way of indicating that there
-                # was an error without causing the sync to fail for
-                # other sources
+            except IOError as e:
+                log.debug('%s: %s', source, e)
                 continue
             finally:
                 # only close files that were opened from the filesystem
                 if isinstance(source, str):
-                    file.close()
+                    format.close(file)
+            
+            one_source_read = True
             
             # return lines only for the first source
             if i == 0:
                 write_lines = lines
         
+        if not one_source_read:
+            log.info('no sources were read')
+        
         return write_lines
     
     def _write(self, source, format, lines=None):
@@ -414,9 +391,9 @@ class ConfigSection(collections.MutableMapping):
         finally:
             # only close files that were opened from the filesystem
             if isinstance(source, str):
-                file.close()
+                format.close(file)
             else:
-                file.flush()
+                format.flush(file)
     
     def _process_sources(self, sources, format):
         sources = sources or self.sources
@@ -432,7 +409,9 @@ class ConfigSection(collections.MutableMapping):
         """
         if not format:
             return self._format
-        elif isinstance(format, str):
+        if isinstance(format, bytes):
+            format = format.decode('ascii')
+        if isinstance(format, str):
             try:
                 cls = Config._formats[format]
             except KeyError as e:
@@ -476,10 +455,10 @@ class ConfigSection(collections.MutableMapping):
     def _keystr(self, key):
         return self._root.sep.join(key)
     
-    def _reset(self):
+    def _reset(self, clean=False):
         if self._value is not NoValue:
             self._value = NoValue
-            self.dirty = True
+            self._dirty = not clean
         if self._default is NoValue:
             self._type = None
     
@@ -560,6 +539,8 @@ class MetaFormat(type):
     :class:`~profig.Config` class.
     """
     def __init__(cls, name, bases, dct):
+        if isinstance(name, bytes):
+            name = name.decode('ascii')
         if name not in ('BaseFormat', 'Format'):
             Config._formats[cls.name] = cls
         return super(MetaFormat, cls).__init__(name, bases, dct)
@@ -592,7 +573,7 @@ class Format(BaseFormat):
         in a subclass."""
         raise NotImplementedError('abstract')
     
-    def open(self, source, mode='rb', *args):
+    def open(self, source, mode='rb'):
         """Returns a file object.
         
         If *source* is a file object, returns *source*. If *mode* is 'w',
@@ -603,16 +584,23 @@ class Format(BaseFormat):
             source = source.decode(self.encoding)
         
         if isinstance(source, str):
+            source = os.path.expanduser(source)
             if self.ensure_dirs is not None and 'w' in mode:
                 # ensure the path exists if any writing is to be done
                 ensure_dirs(os.path.dirname(source), self.ensure_dirs)
-            return io.open(source, mode, *args)
+            return io.open(source, mode)
         else:
             source.seek(0)
             if 'w' in mode:
                 source.truncate()
             return source
     
+    def close(self, file):
+        file.close()
+    
+    def flush(self, file):
+        file.flush()
+    
     def _read_error(self, file, lineno=None, text='', message=''):
         if self.read_errors != 'ignore':
             name = file.name if hasattr(file, 'name') else '<???>'
@@ -710,8 +698,9 @@ class IniFormat(Format):
         # file has been read. assign the values
         for key, value in values.items():
             section = cfg.section(key)
-            if not section.dirty and value is not None:
-                section.set_value(value, adapt=False)
+            if not section._dirty and value is not None:
+                section.convert(value)
+                section._dirty = False
             if key in comments:
                 section.comment = comments[key]
         
@@ -731,7 +720,7 @@ class IniFormat(Format):
         if section.parent is section.root:
             # header section
             if section.valid:
-                value = section.value(convert=False)
+                value = section.adapt()
                 write('[{}]'.format(section.name))
                 file.write(self.delimeter)
                 if isinstance(value, bytes):
@@ -746,7 +735,7 @@ class IniFormat(Format):
             # value section
             cfg = self.config
             key = cfg._keystr(cfg._make_key(section.key)[1:])
-            value = section.value(convert=False)
+            value = section.adapt()
             write(key)
             file.write(self.delimeter)
             if isinstance(value, bytes):
@@ -755,7 +744,7 @@ class IniFormat(Format):
                 write(value)
             write('\n')
         
-        section.dirty = False
+        section._dirty = False
     
     def write(self, file, lines=None):
         cfg = self.config
@@ -808,6 +797,95 @@ class IniFormat(Format):
             seen.add(section.key)
             first = False
 
+if WIN:
+    class RegistryFormat(Format):
+        name = 'registry'
+        base_key = winreg.HKEY_CURRENT_USER
+        
+        def read(self, key, section=None):
+            section = section if section is not None else self.config
+            n_subkeys, n_values, t = winreg.QueryInfoKey(key)
+            
+            # read values from this subkey
+            for i in range(n_values):
+                name, value, type = winreg.EnumValue(key, i)
+                subsection = section.section(name)
+                subsection.set_value(value)
+                subsection._dirty = False
+            
+            # read values from next subkeys
+            for i in range(n_subkeys):
+                name = winreg.EnumKey(key, i)
+                subkey = winreg.OpenKeyEx(key, name)
+                subsection = section.section(name)
+                self.read(subkey, subsection)
+        
+        def write(self, key, context=None):
+            cfg = self.config
+            for section in cfg.sections(recurse=True, only_valid=True):
+                # determine the registry key/name
+                section_key = cfg._make_key(section.key)
+                if section.has_children:
+                    rkey = ntpath.sep.join(section_key)
+                    name = ''
+                else:
+                    rkey = ntpath.sep.join(section_key[:-1])
+                    name = section_key[-1]
+                
+                # write the value
+                subkey = winreg.CreateKeyEx(key, rkey)
+                value = section.value()
+                type = self._get_type(value)
+                winreg.SetValueEx(subkey, name, 0, type, value)
+        
+        def open(self, source, mode='rb'):
+            if 'r' in mode:
+                try:
+                    return winreg.OpenKeyEx(self.base_key, source)
+                except WindowsError as e:
+                    raise IOError(e)
+            elif 'w' in mode:
+                return winreg.CreateKeyEx(self.base_key, source)
+            else:
+                raise ValueError('invalid mode: {}'.format(mode))
+        
+        def close(self, key):
+            winreg.CloseKey(key)
+        
+        def flush(self, key):
+            winreg.FlushKey(key)
+        
+        def delete(self, key):
+            """Deletes all keys and values recursively from *key*."""
+            for subkey in list(self.all_keys(key)):
+                winreg.DeleteKey(subkey, '')
+        
+        def all_keys(self, key):
+            """Generates all keys descending from *key*.
+            
+            The deepest is returned first, then the rest, all the way
+            back to *key*
+            """
+            n_subkeys, n_values, t = winreg.QueryInfoKey(key)
+            for i in range(n_subkeys):
+                name = winreg.EnumKey(key, i)
+                subkey = winreg.OpenKeyEx(key, name)
+                for subsubkey in self.all_keys(subkey):
+                    yield subsubkey
+            
+            yield key
+        
+        def _get_type(self, value):
+            if isinstance(value, str):
+                return winreg.REG_SZ
+            elif isinstance(value, bytes):
+                return winreg.REG_BINARY
+            elif isinstance(value, int):
+                return winreg.REG_DWORD
+            else:
+                err = 'type not supported by this format: {}'
+                raise ValueError(err.format(_type(value)))
+
 ## Config Errors ##
 
 class ConfigError(Exception):
